# -*- coding: utf-8 -*-

import itertools
import numpy as np


def entropy(*X):
    """
    Calculate the entropy of a variable, or joint entropy of several variables.

    Parameters
    ----------
    X : array, or list of arrays
        Variable or variables to compute entropy/joint entropy on

    Notes
    -----
    This function can be used to calculate the entropy of a single variable
    (provided as a single input) or to calculate the joint entropy between two
    variables (provided as a series of inputs)
    """
    n_instances = len(X[0])
    H = 0
    for classes in itertools.product(*[set(x) for x in X]):
        v = np.array([True] * n_instances)
        for predictions, c in zip(X, classes):
            v = np.logical_and(v, predictions == c)
        p = np.mean(v)
        H += -p * np.log2(p) if p > 0 else 0
    return H


def conditional_entropy(x, y):
    """
    The conditional entropy H(X|Y) = H(Y,X) - H(Y). X conditioned on Y
    """
    H_y = entropy(y)
    H_yx = entropy(y, x)
    return H_yx - H_y


def mutual_information(x, y):
    """
    The mutual information between two variables

    MI(X, Y) = H(X) + H(Y) - H(X | Y)

    Parameters
    ----------
    x, y : array

    Returns
    -------
    array : mutual information between x and y
    """
    H_x = entropy(x)
    H_y = entropy(y)
    H_xy = entropy(x, y)
    return H_x + H_y - H_xy


def entropy_cc(x, y):
    """
    The entropy correlation coefficient:

    p(H) = sqrt(MI(X, Y) / 0.5 * (H(X) + H(Y)))
    """
    H_x = entropy(x)
    H_y = entropy(y)
    I_xy = mutual_information(y, x)
    return np.sqrt(I_xy / (0.5 * (H_x + H_y)))


def transfer_entropy(x, y, lag=1):
    """
    Transfer entropy for two given signals.

    Parameters
    ----------
      x : array
        source
      y : array
        target
      lag : int

    Returns
    -------
    array : Transfer entropy from x to y
    """
    # Future of i
    Fi = np.roll(x, -lag)
    # Past of i
    Pi = x
    # Past of j
    Pj = y

    # Transfer entropy
    Inf_from_Pi_to_Fi = conditional_entropy(Fi, Pi)

    # Same as cond_entropy(Fi, Pi_Pj)
    H_y = entropy(Pi, Pj)
    H_yx = entropy(Fi, Pj, Pi)
    Inf_from_Pi_Pj_to_Fi = H_yx - H_y

    TE_from_j_to_i = Inf_from_Pi_to_Fi - Inf_from_Pi_Pj_to_Fi

    return TE_from_j_to_i



import numpy as np

from nitime import descriptors as desc
from nitime import timeseries as ts
from nitime import algorithms as tsa

# To support older versions of numpy that don't have tril_indices:
from nitime.index_utils import tril_indices

from .base import BaseAnalyzer


class CorrelationAnalyzer(BaseAnalyzer):
    """Analyzer object for correlation analysis. Has the same API as the
    CoherenceAnalyzer"""

    def __init__(self, input=None):
        """
        Parameters
        ----------

        input : TimeSeries object
           Containing the data to analyze.

        Examples
        --------
        >>> t1 = ts.TimeSeries(data = np.sin(np.arange(0,
        ...                    10*np.pi,10*np.pi/100)).reshape(2,50),
        ...                                      sampling_rate=np.pi)
        >>> c1 = CorrelationAnalyzer(t1)
        >>> c1 = CorrelationAnalyzer(t1)
        >>> c1.corrcoef
        array([[ 1., -1.],
               [-1.,  1.]])
        >>> c1.xcorr.sampling_rate  # doctest: +ELLIPSIS
        3.141592653... Hz
        >>> c1.xcorr.t0  # doctest: +ELLIPSIS
        -15.91549430915... s

        """

        BaseAnalyzer.__init__(self, input)

    @desc.setattr_on_read
    def corrcoef(self):
        """The correlation coefficient between every pairwise combination of
        time-series contained in the object"""
        return np.corrcoef(self.input.data)

    @desc.setattr_on_read
    def xcorr(self):
        """The cross-correlation between every pairwise combination time-series
        in the object. Uses np.correlation('full').

        Returns
        -------

        TimeSeries : the time-dependent cross-correlation, with zero-lag
        at time=0

        """
        tseries_length = self.input.data.shape[0]
        t_points = self.input.data.shape[-1]
        xcorr = np.zeros((tseries_length,
                          tseries_length,
                          t_points * 2 - 1))
        data = self.input.data
        for i in range(tseries_length):
            data_i = data[i]
            for j in range(i, tseries_length):
                xcorr[i, j] = np.correlate(data_i,
                                          data[j],
                                          mode='full')

        idx = tril_indices(tseries_length, -1)
        xcorr[idx[0], idx[1], ...] = xcorr[idx[1], idx[0], ...]

        return ts.TimeSeries(xcorr,
                             sampling_interval=self.input.sampling_interval,
                             t0=-self.input.sampling_interval * t_points)

    @desc.setattr_on_read
    def xcorr_norm(self):
        """The cross-correlation between every pairwise combination time-series
        in the object, where the zero lag correlation is normalized to be equal
        to the correlation coefficient between the time-series

        Returns
        -------

        TimeSeries : A TimeSeries object
            the time-dependent cross-correlation, with zero-lag at time=0

        """

        tseries_length = self.input.data.shape[0]
        t_points = self.input.data.shape[-1]
        xcorr = np.zeros((tseries_length,
                          tseries_length,
                          t_points * 2 - 1))
        data = self.input.data
        for i in range(tseries_length):
            data_i = data[i]
            for j in range(i, tseries_length):
                xcorr[i, j] = np.correlate(data_i,
                                          data[j],
                                          mode='full')
                xcorr[i, j] /= (xcorr[i, j, t_points])
                xcorr[i, j] *= self.corrcoef[i, j]

        idx = tril_indices(tseries_length, -1)
        xcorr[idx[0], idx[1], ...] = xcorr[idx[1], idx[0], ...]

        return ts.TimeSeries(xcorr,
                             sampling_interval=self.input.sampling_interval,
                             t0=-self.input.sampling_interval * t_points)


class SeedCorrelationAnalyzer(object):
    """
    This analyzer takes two time-series. The first is designated as a
    time-series of seeds. The other is designated as a time-series of targets.
    The analyzer performs a correlation analysis between each of the channels
    in the seed time-series and *all* of the channels in the target
    time-series.

    """
    def __init__(self, seed_time_series=None, target_time_series=None):
        """
        Parameters
        ----------

        seed_time_series : a TimeSeries object

        target_time_series : a TimeSeries object

        """
        self.seed = seed_time_series
        self.target = target_time_series

    @desc.setattr_on_read
    def corrcoef(self):

        #If there is more than one channel in the seed time-series:
        if len(self.seed.shape) > 1:

            # Preallocate results
            Cxy = np.empty((self.seed.data.shape[0],
                            self.target.data.shape[0]), dtype=float)

            for seed_idx, this_seed in enumerate(self.seed.data):

                Cxy[seed_idx] = tsa.seed_corrcoef(this_seed, self.target.data)

        #In the case where there is only one channel in the seed time-series:
        else:
            Cxy = tsa.seed_corrcoef(self.seed.data, self.target.data)

        return Cxy.squeeze()



"""

.. _filter-fmri:

===================================
Filtering and normalizing fMRI data
===================================

Filtering fMRI data is very important. The time-series usually collected in
fMRI contain a broad-band signal. However, physilogically relevant signals are
thought to be present in only particular parts of the spectrum. For this
reason, filtering operations, such as detrending, are a common pre-processing
step in analysis of fMRI data analysis. In addition, for many fMRI analyses, it
is important to normalize the data in each voxel. This is because data may
differ between different voxels for 'uninteresting' reasons, such as local
blood-flow differences and signal amplitude differences due to the distance
from the receive coil. In the following, we will demonstrate usage of nitimes
analyzer objects for spectral estimation, filtering and normalization on fMRI
data.


We start by importing the needed modules. First modules from the standard lib
and from 3rd parties:

"""

import os

import numpy as np
import matplotlib.pyplot as plt


"""

Next, the particular nitime classes we will be using in this example:

"""

import nitime

# Import the time-series objects:
from nitime.timeseries import TimeSeries

# Import the analysis objects:
from nitime.analysis import SpectralAnalyzer, FilterAnalyzer, NormalizationAnalyzer

"""

For starters, let's analyze data that has been preprocessed and is extracted
into indivudal ROIs. This is the same data used in :ref:`multi-taper-coh` and
in :ref:`resting-state` (see these examples for details).

We start by setting the TR and reading the data from the CSV table into which
the data was saved:

"""

TR = 1.89

data_path = os.path.join(nitime.__path__[0], 'data')

data = np.loadtxt(os.path.join(data_path, 'fmri_timeseries.csv'),
                  skiprows=1, delimiter=',')

# Extract ROI information from the csv file headers:
roi_names = np.array(data_rec.dtype.names)

# This is the number of samples in each ROI:
n_samples = data_rec.shape[0]

# Make an empty container for the data
data = np.zeros((len(roi_names), n_samples))

# Insert the data from each ROI into a row in the data:
for n_idx, roi in enumerate(roi_names):
    data[n_idx] = data_rec[roi]

# Initialize TimeSeries object:
T = TimeSeries(data, sampling_interval=TR)
T.metadata['roi'] = roi_names


"""

We will start, by examining the spectrum of the original data, before
filtering. We do this by initializing a SpectralAnalyzer for the original data:

"""

S_original = SpectralAnalyzer(T)

# Initialize a figure to put the results into:
fig01 = plt.figure()
ax01 = fig01.add_subplot(1, 1, 1)


"""

The spectral analyzer has several different methods of spectral analysis,
however the all have a common API. This means that for all of them the output
is a tuple. The first item in the tuple are the central frequencies of the
frequency bins in the spectrum and the second item in the tuple are the
magnitude of the spectrum in that frequency bin. For the purpose of this
example, we will only plot the data from the 10th ROI (by indexing into the
spectra). We compare all the methods of spectral estimation by plotting them
together:

"""

ax01.plot(S_original.psd[0],
          S_original.psd[1][9],
          label='Welch PSD')

ax01.plot(S_original.spectrum_fourier[0],
          np.abs(S_original.spectrum_fourier[1][9]),
          label='FFT')

ax01.plot(S_original.periodogram[0],
          S_original.periodogram[1][9],
          label='Periodogram')

ax01.plot(S_original.spectrum_multi_taper[0],
          S_original.spectrum_multi_taper[1][9],
          label='Multi-taper')

ax01.set_xlabel('Frequency (Hz)')
ax01.set_ylabel('Power')

ax01.legend()


"""

.. image:: fig/filtering_fmri_01.png


Notice that, for this data, simply extracting a FFT is hardly informative (the
reasons for that are explained in :ref:`multi-taper-psd`). On the other hand,
the other methods provide different granularity of information, traded-off with
the robustness of the estimation. The cadillac of spectral estimates is the
multi-taper estimation, which provides both robustness and granularity, but
notice that this estimate requires more computation than other estimates
(certainly more estimates than the FFT).

We note that a lot of the power in the fMRI data seems to be concentrated in
frequencies below 0.02 Hz. These extremely low fluctuations in signal are often
considered to be 'noise', rather than reflecting neural processing. In
addition, there is a broad distribution of power up to the Nyquist
frequency. However, some estimates of the hemodynamic response suggest that
information above 0.15 could not reflect the slow filtering of neural response
to the BOLD response measured in fMRI. Thus, it would be advantageous to remove
fluctuations below 0.02 and above 0.15 Hz from the data. Next, we proceed to
filter the data into this range, using different methods.

We start by initializing a FilterAnalyzer. This is initialized with the
time-series containing the data and with the upper and lower bounds of the
range into which we wish to filter (in Hz):

"""

F = FilterAnalyzer(T, ub=0.15, lb=0.02)

# Initialize a figure to display the results:
fig02 = plt.figure()
ax02 = fig02.add_subplot(1, 1, 1)

# Plot the original, unfiltered data:
ax02.plot(F.data[0], label='unfiltered')

"""

As with the SpectralAnalyzer, there is a common API for the different methods
used for filtering. We use the following methods:

- Boxcar filter: The time-series is convolved with a box-car function of the
  right length to smooth the data to such an extent that the frequencies higher
  than represented by the length of this box-car function are no longer present
  in the smoothed version of the time-series. This functions as a low-pass filter. The
  data can then be high-pass filtered by subtracting this version of the data
  from the original. For a band-pass filter, both of these operations are done.

"""

ax02.plot(F.filtered_boxcar.data[0], label='Boxcar filter')

"""

- FIR filter: A digital filter with a finite impulse response. These filters
  have an order of 64 per default, but that can be adjusted by setting the key
  word argument 'filt_order', passed to initialize the FilterAnalyzer. For
  FIR filtering, :mod:`nitime` uses a Hamming window filter, but this can also
  be changed by setting the key word argument 'fir_win'.
  As with the boxcar filter, if band-pass filtering is required, a low-pass
  filter is applied and then a high-pass filter is applied to the resulting
  time-series.

"""

ax02.plot(F.fir.data[0], label='FIR')

"""

- IIR filter: A digital filter with an infinite impulse response function. Per
  default an elliptic filter is used here, but this can be changed, by setting
  the 'iir_type' key word argument used when initializing the FilterAnalyzer.

For both FIR filters and IIR filters, :func:`scipy.signal.filtfilt` is used in
order to achieve zero phase delay filtering.

"""

ax02.plot(F.iir.data[0], label='IIR')

"""

- Fourier filter: this is a quick and dirty filter. The data is FFT-ed into the
  frequency domain. The power in the unwanted frequency bins is removed (by
  replacing the power in these bins with zero) and the data is IFFT-ed back
  into the time-domain.

"""

ax02.plot(F.filtered_fourier.data[0], label='Fourier')
ax02.legend()
ax02.set_xlabel('Time (TR)')
ax02.set_ylabel('Signal amplitude (a.u.)')

"""

.. image:: fig/filtering_fmri_02.png


Examining the resulting time-series closely reveals that large fluctuations in
very slow frequencies have been removed, but also small fluctuations in high
frequencies have been attenuated through filtering.

Comparing the resulting spectra of these different filters shows the various
trade-offs of each filtering method, including the fidelity with which the
original spectrum is replicated within the pass-band and the amount of
attenuation within the stop-bands.

We can do that by initializng a SpectralAnalyzer for each one of the filtered
time-series resulting from the above operation and plotting their spectra. For
ease of compariso, we only plot the spectra using the multi-taper spectral
estimation. At the level of granularity provided by this method, the differences
between the methods are emphasized:

"""

S_fourier = SpectralAnalyzer(F.filtered_fourier)
S_boxcar = SpectralAnalyzer(F.filtered_boxcar)
S_fir = SpectralAnalyzer(F.fir)
S_iir = SpectralAnalyzer(F.iir)

fig03 = plt.figure()
ax03 = fig03.add_subplot(1, 1, 1)

ax03.plot(S_original.spectrum_multi_taper[0],
          S_original.spectrum_multi_taper[1][9],
          label='Original')

ax03.plot(S_fourier.spectrum_multi_taper[0],
          S_fourier.spectrum_multi_taper[1][9],
          label='Fourier')

ax03.plot(S_boxcar.spectrum_multi_taper[0],
          S_boxcar.spectrum_multi_taper[1][9],
          label='Boxcar')

ax03.plot(S_fir.spectrum_multi_taper[0],
          S_fir.spectrum_multi_taper[1][9],
          label='FIR')

ax03.plot(S_iir.spectrum_multi_taper[0],
          S_iir.spectrum_multi_taper[1][9],
          label='IIR')

ax03.legend()


"""

.. image:: fig/filtering_fmri_03.png


Next, we turn to normalize the filtered data. This can be done in one of two
methods:

- Percent change: the data in each voxel is normalized as percent signal
  change, relative to the mean BOLD signal in the voxel

- Z score: The data in each voxel is normalized to have 0 mean and a standard
  deviation of 1.

We will use the filtered data, in order to demonstrate how the output of one
analyzer can be used as the input to the other:

"""

fig04 = plt.figure()
ax04 = fig04.add_subplot(1, 1, 1)

ax04.plot(NormalizationAnalyzer(F.fir).percent_change.data[0], label='% change')
ax04.plot(NormalizationAnalyzer(F.fir).z_score.data[0], label='Z score')
ax04.legend()
ax04.set_xlabel('Time (TR)')
ax04.set_ylabel('Amplitude (% change or Z-score)')

"""

.. image:: fig/filtering_fmri_04.png


Notice that the same methods of filtering and normalization can be applied to
fMRI data, upon reading it from a nifti file, using :mod:`nitime.fmri.io`.

We demonstrate that in what follows.[Notice that nibabel
(http://nipy.org/nibabel) is required in order to run the following
examples. An error will be thrown if nibabel is not installed]

"""

try:
    from nibabel import load
except ImportError:
    raise ImportError('You need nibabel (http:/nipy.org/nibabel/) in order to run this example')

import nitime.fmri.io as io

"""

We define the TR of the analysis and the frequency band of interest:

"""

TR = 1.35
f_lb = 0.02
f_ub = 0.15


"""

An fMRI data file with some fMRI data is shipped as part of the distribution,
the following line will find the path to this data on the specific computer:

"""

data_file_path = test_dir_path = os.path.join(nitime.__path__[0],
                                              'data')

fmri_file = os.path.join(data_file_path, 'fmri1.nii.gz')


"""

Read in the dimensions of the data, using nibabel:

"""

fmri_data = load(fmri_file)
volume_shape = fmri_data.shape[:-1]
coords = list(np.ndindex(volume_shape))
coords = np.array(coords).T


"""

We use :mod:`nitime.fmri.io` in order to generate a TimeSeries object from spatial
coordinates in the data file. Notice that normalization method is provided as a
string input to the keyword argument 'normalize' and the filter and its
properties are provided as a dict to the keyword argument 'filter':

"""

T_unfiltered = io.time_series_from_file(fmri_file,
                                        coords,
                                        TR=TR,
                                        normalize='percent')

T_fir = io.time_series_from_file(fmri_file,
                              coords,
                              TR=TR,
                              normalize='percent',
                              filter=dict(lb=f_lb,
                                          ub=f_ub,
                                          method='fir',
                                          filt_order=10))

T_iir = io.time_series_from_file(fmri_file,
                              coords,
                              TR=TR,
                              normalize='percent',
                              filter=dict(lb=f_lb,
                                          ub=f_ub,
                                          method='iir',
                                          filt_order=10))

T_boxcar = io.time_series_from_file(fmri_file,
                              coords,
                              TR=TR,
                              normalize='percent',
                              filter=dict(lb=f_lb,
                                          ub=f_ub,
                                          method='boxcar',
                                          filt_order=10))

fig05 = plt.figure()
ax05 = fig05.add_subplot(1, 1, 1)
S_unfiltered = SpectralAnalyzer(T_unfiltered).spectrum_multi_taper
S_fir = SpectralAnalyzer(T_fir).spectrum_multi_taper
S_iir = SpectralAnalyzer(T_iir).spectrum_multi_taper
S_boxcar = SpectralAnalyzer(T_boxcar).spectrum_multi_taper

random_voxel = np.random.randint(0, np.prod(volume_shape))

ax05.plot(S_unfiltered[0], S_unfiltered[1][random_voxel], label='Unfiltered')
ax05.plot(S_fir[0], S_fir[1][random_voxel], label='FIR filtered')
ax05.plot(S_iir[0], S_iir[1][random_voxel], label='IIR filtered')
ax05.plot(S_boxcar[0], S_boxcar[1][random_voxel], label='Boxcar filtered')
ax05.legend()

"""

.. image:: fig/filtering_fmri_05.png


Notice that though the boxcar filter doesn't usually do an amazing job with
long time-series and IIR/FIR filters seem to be superior in those cases, in
this example, where the time-series is much shorter, it sometimes does a
relatively decent job.

We call plt.show() in order to display the figure:

"""

plt.show()



""" This module provides lazy import functionality to improve the import
performance of nitime. For example, some parts of nitime leverage and import
matplotlib, which is quite a big package, yet most of the nitime code does not
depend on matplotlib. By lazily-loading a module, we defer the overhead of
importing it until the first time it is actually used, thereby speeding up
nitime imports.

A generic :class:`LazyImport` class is implemented which takes the module name
as a parameter, and acts as a proxy for that module, importing it only when
the module is used, but effectively acting as the module in every other way
(including inside IPython with respect to introspection and tab completion)
with the *exception* of reload() - reloading a :class:`LazyImport` raises an
:class:`ImportError`.

Commonly used nitime lazy imports are also defined in :mod:`nitime.lazy`, so
they can be reused throughout nitime.
"""
import sys
import types

# This flag only has affect on this module's import, and if it is set to True,
# LazyImports are performed immediately. Note: this flag is currently here
# only for debugging purposes and must be set directly in the source code,
# since nitime.lazy imports this module, and nitime.lazy is used throughout
# nitime, importing nitime will import this module.
disable_lazy_imports = False

if 'sphinx' in sys.modules:
    disable_lazy_imports = True

class LazyImport(types.ModuleType):
    """
    This class takes the module name as a parameter, and acts as a proxy for
    that module, importing it only when the module is used, but effectively
    acting as the module in every other way (including inside IPython with
    respect to introspection and tab completion) with the *exception* of
    reload()- reloading a :class:`LazyImport` raises an :class:`ImportError`.

    >>> numpy = LazyImport('numpy')

    No import happens on the above line, until we do something like call an
    ``numpy`` method or try to do tab completion or introspection on ``numpy``
    in IPython.

    >>> numpy
    <module 'numpy' will be lazily loaded>

    Now the :class:`LazyImport` will do an actual import, and call the hypot
    function of the imported module.

    >>> numpy.diff([1969, 2011])
    array([42])
    """
    def __getattribute__(self,x):
        # This method will be called only once, since we'll change
        # self.__class__ to LoadedLazyImport, and __getattribute__ will point
        # to module.__getattribute__
        name = object.__getattribute__(self,'__name__')
        __import__(name)
        # if name above is 'package.foo.bar', package is returned, the docs
        # recommend that in order to get back the full thing, that we import
        # and then lookup the full name is sys.modules, see:
        # http://docs.python.org/library/functions.html#__import__
        module = sys.modules[name]
        # Now that we've done the import, cutout the middleman and make self
        # act as the imported module
        class LoadedLazyImport(types.ModuleType):
            __getattribute__ = module.__getattribute__
            __repr__ = module.__repr__
        object.__setattr__(self,'__class__', LoadedLazyImport)
        # The next line will make "reload(l)" a silent no-op
        #sys.modules[name] = self
        return module.__getattribute__(x)
    def __repr__(self):
        return "<module '%s' will be lazily loaded>" %\
                object.__getattribute__(self,'__name__')

if disable_lazy_imports:
    lazy_doc = """:class:`LazyImports` have been globally disabled.
    Please modify ``disable_lazy_imports`` boolean variable in
    :mod:`nitime.lazyimports` in order to leverage lazy loading of modules.
    """
    if 'sphinx' in sys.modules:
        lazy_doc = """
                   WARNING: To get Sphinx documentation to build we disable
                   LazyImports, which makes Sphinx incorrectly report this
                   class as having a base class of object. In reality,
                   :class:`LazyImport`'s base class is
                   :class:`types.ModuleType`.
                   """
        lazy_doc += LazyImport.__doc__
    class LazyImport(object):
        __doc__ = lazy_doc
        def __init__(self, x):
            __import__(x)
            self.module = sys.modules[x]
        def __getattr__(self, x):
            return self.module.__getattribute__(x)



"""Base classes for generic time series analysis.

The classes implemented here are meant to provide fairly basic objects for
managing time series data.  They should serve mainly as data containers, with
only minimal algorithmic functionality.

In the timeseries subpackage, there is a separate library of algorithms, and
the classes defined here mostly delegate any computational facilities they may
have to that library.

Over time, it is OK to add increasingly functionally rich classes, but only
after their design is well proven in real-world use.

"""
#-----------------------------------------------------------------------------
# Public interface
#-----------------------------------------------------------------------------
__all__ = ['time_unit_conversion',
           'TimeSeriesInterface',
           'TimeSeries',
           'TimeInterface',
           'UniformTime',
           'TimeArray',
           'Epochs',
           'Events'
           ]
#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import numpy as np

# Our own
from nitime import descriptors as desc

try:
    _NP_2 = int(np.__version__.split(".")[0]) >= 2
except Exception:
    _NP_2 = True

#-----------------------------------------------------------------------------
# Module globals
#-----------------------------------------------------------------------------

# These are the valid names for time units, taken from the Numpy date/time
# types specification document.  They conform to SI nomenclature where
# applicable.

# Most uses of this are membership checks, so we make a set for fast
# validation.  But we create them first as a list so we can print an ordered
# and easy to read error message.

time_unit_conversion = {
                        'ps': 1,  # picosecond
                        'ns': 10 ** 3,  # nanosecond
                        'us': 10 ** 6,  # microsecond
                        'ms': 10 ** 9,  # millisecond
                        's': 10 ** 12,   # second
                         None: 10 ** 12,  # The default is seconds (when
                                        # constructor doesn't get any
                                        # input, it defaults to None)
                        'm': 60 * 10 ** 12,   # minute
                        'h': 3600 * 10 ** 12,   # hour
                        'D': 24 * 3600 * 10 ** 12,   # day
                        'W': 7 * 24 * 3600 * 10 ** 12,  # week
                                                        # (not an SI unit)
                        }

# The basic resolution:
base_unit = 'ps'


#-----------------------------------------------------------------------------
# Class declarations
#-----------------------------------------------------------------------------

# Time:
class TimeInterface(object):
    """ The minimal object interface for time representations

    This should be thought of as an abstract base class. """

    time_unit = None


def get_time_unit(obj):
    """
    Extract the time unit of the object. If it is an iterable, get the time
    unit of the first element.
    """

    # If this is a Time object, no problem:
    if isinstance(obj, TimeInterface):
        return obj.time_unit

    # Otherwise, if it is iterable, we recurse on it:
    try:
        it = iter(obj)
    except TypeError:
        return None
    else:
        return get_time_unit(next(it))


class TimeArray(np.ndarray, TimeInterface):
    """Base-class for time representations, implementing the TimeInterface"""
    def __new__(cls, data, time_unit=None, copy=True):
        """
        Parameters
        ----------
        data : 1-d array or `TimeArray` class instance
            Time points

        time_unit : str, optional
            The time-unit to use. This should be one of the keys of the
            `time_unit_conversion` dict from the :mod:`timeseries` module,
            which are SI units of time. Default: 's'

        copy : bool, optional
            Whether to create this instance by copy of a. If False,
            a copy will not be forced but might still be required depending
            on the data array.

        Note
        ----
        If the 'copy' input is set to False, input must be either a `TimeArray`
        class instance, or an int64 array in the base unit of the module
        (which, unless you change it, is picoseconds)


        """

        # Check that the time units provided are sensible:
        if time_unit not in time_unit_conversion:
            raise ValueError('Invalid time unit %s, must be one of %s' %
                             (time_unit, time_unit_conversion.keys()))

        # Get the conversion factor from the input:
        conv_fac = time_unit_conversion[time_unit]

        # Call get_time_unit to pull the time_unit out from inside:
        data_time_unit = get_time_unit(data)
        # If it has a time unit, you should not convert the values to
        # base_unit, because they are already in that:
        if data_time_unit is not None:
            conv_fac = 1

        # We check whether the data has a time-unit somewhere inside (for
        # example, if it is a list of TimeArray objects):
        if time_unit is None:
            time_unit = data_time_unit

        # We can only honor the copy flag in a very narrow set of cases
        # if data is already a TimeArray or if data is an ndarray with
        # dtype=int64
        if copy == False:
            if not getattr(data, 'dtype', None) == np.int64:
                e_s = 'When copy flag is set to False, must provide a'
                e_s += 'TimeArray in object, or int64 times, in %s' % base_unit
                raise ValueError(e_s)

            time = np.asarray(data)
        else:
            if isinstance(data, TimeInterface):
                time = data.copy()
            else:
                data_arr = np.asarray(data)
                if issubclass(data_arr.dtype.type, np.integer):
                    # If this is an array of integers, cast to 64 bit integer
                    # and convert to the base_unit.
                    #XXX This will fail when even 64 bit is not large enough to
                    # avoid wrap-around (When you try to make more than 10**6
                    # seconds). XXX this should be mentioned in the docstring
                    time = data_arr.astype(np.int64) * conv_fac
                else:
                    # Otherwise: first convert, round and then cast to 64
                    time = (data_arr * conv_fac).round().astype(np.int64)

        # Make sure you have an array on your hands (for example, if you input
        # an integer, you might have reverted to an integer when multiplying
        # with the conversion factor:
        time = np.asarray(time).view(cls)

        # Make sure time is one-dimensional or 0-d
        if time.ndim > 1:
            raise ValueError('TimeArray can only be one-dimensional or 0-d')

        if time_unit is None:
            time_unit = 's'

        time.time_unit = time_unit
        time._conversion_factor = time_unit_conversion[time_unit]
        return time

    def __array_wrap__(self, out_arr, context=None, return_scalar=False):
        # When doing comparisons between TimeArrays, make sure that you return
        # a boolean array, not a time array:
        if out_arr.dtype == bool:
            ret = np.asarray(out_arr)
            if return_scalar:
                ret = ret[()]
            return ret
        else:
            return np.ndarray.__array_wrap__(self, out_arr, context, return_scalar)

    def __array_finalize__(self, obj):
        """XXX """
        # Make sure that the TimeArray has the time units set (and not equal to
        # None):
        if not hasattr(self, 'time_unit') or self.time_unit is None:
            if hasattr(obj, 'time_unit'):  # looks like view cast
                self.time_unit = obj.time_unit
            else:
                self.time_unit = 's'

        # Make sure that the conversion factor is set properly:
        if not hasattr(self, '_conversion_factor'):
            if hasattr(obj, '_conversion_factor'):
                self._conversion_factor = obj._conversion_factor
            else:
                self._conversion_factor = time_unit_conversion[self.time_unit]

    def __repr__(self):
        """Pass it through the conversion factor"""

        # If the input is a single int/float (with no shape) return a 'scalar'
        # time-point:
        if self.shape == ():
            return "%r %s" % (int(self) / float(self._conversion_factor),
                           self.time_unit)
        # Otherwise, return the TimeArray representation:
        else:
            return np.ndarray.__repr__(self / float(self._conversion_factor)
             )[:-1] + ", time_unit='%s')" % self.time_unit

    def __str__(self):
        """Return a nice string representation of this TimeArray"""
        return self.__repr__()

    def __getitem__(self, key):
        # return scalar TimeArray in case key is integer
        if isinstance(key, (int, np.int64, np.int32)):
            return self[[key]].reshape(())
        elif isinstance(key, float):
            return self.at(key)
        elif isinstance(key, Epochs):
            return self.during(key)
        else:
            return np.ndarray.__getitem__(self, key)

    def __setitem__(self, key, val):
        # look at the units - convert the values to what they need to be (in
        # the base_unit) and then delegate to the ndarray.__setitem__
        if not hasattr(val, '_conversion_factor'):
            val *= self._conversion_factor
        return np.ndarray.__setitem__(self, key, val)

    def _convert_if_needed(self,val):
        if not hasattr(val, '_conversion_factor'):
            val = np.asarray(val)
            if getattr(val, 'dtype', None) == np.int32:
                # we'll overflow if val's dtype is np.int32
                val = np.array(val, dtype=np.int64)
            val *= self._conversion_factor
        return val

    def __add__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__add__(self, val)

    def __sub__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__sub__(self, val)

    def __radd__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__radd__(self, val)

    def __rsub__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__rsub__(self, val)

    def __lt__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__lt__(self, val)

    def __gt__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__gt__(self, val)

    def __le__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__le__(self, val)

    def __ge__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__ge__(self,val)

    def __eq__(self, val):
        val = self._convert_if_needed(val)
        return np.ndarray.__eq__(self,val)

    def min(self, *args, **kwargs):
        ret = TimeArray(np.ndarray.min(self, *args, **kwargs),
            time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def max(self, *args,**kwargs):
        ret = TimeArray(np.ndarray.max(self, *args, **kwargs),
            time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def mean(self, *args, **kwargs):
        ret = TimeArray(np.ndarray.mean(self, *args, **kwargs),
            time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def ptp(self, *args, **kwargs):
        if _NP_2:
            ptp = np.ptp
        else:
            ptp = np.ndarray.ptp
        ret = TimeArray(ptp(self, *args, **kwargs),
                        time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def sum(self, *args,**kwargs):
        ret = TimeArray(np.ndarray.sum(self, *args, **kwargs),
                        time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def prod(self, *args, **kwargs):
        e_s = "Product computation changes TimeArray units"
        raise NotImplementedError(e_s)

    def var(self, *args, **kwargs):
        e_s = "Variance computation changes TimeArray units"
        raise NotImplementedError(e_s)

    def std(self, *args, **kwargs):
        """Returns the standard deviation of this TimeArray (with time units)
        for detailed information, see numpy.std()
        """
        ret = TimeArray(np.ndarray.std(self, *args, **kwargs),
                        time_unit=base_unit)
        ret.convert_unit(self.time_unit)
        return ret

    def index_at(self, t, tol=None, mode='closest'):
        """ Returns the integer indices that corresponds to the time t

        The returned indices depend on both `tol` and `mode`.  The `tol`
        parameter specifies how close the given time must be to those present
        in the array to give a match, when `mode` is `closest`.  The default
        tolerance is 1 `base_unit` (by default, picoseconds).  If you specify
        the tolerance as 0, then only *exact* matches are allowed, be careful
        in this case of possible problems due to floating point roundoff error
        in your time specification.

        When mode is `before` or `after`, the tolerance is completely ignored.
        In this case, either the largest time equal or *before* the given `t`
        or the earliest time equal or *after* the given `t` is returned.

        Parameters
        ----------
        t : time-like
          Anything that is valid input for a TimeArray constructor.
        tol : time-like, optional
          Tolerance, specified in the time units of this TimeArray.
        mode : string
          One of ['closest', 'before', 'after'].

        Returns
        -------
        idx : The array with all the indices where the condition is met.
          """
        if not np.iterable(t):
            t = [t]
        t_e = TimeArray(t, time_unit=self.time_unit)
        if mode == 'closest':
            return self._index_closest(t_e, tol)
        elif mode == 'before':
            return self._index_before(t_e)
        elif mode == 'after':
            return self._index_after(t_e)
        else:
            raise ValueError('Invalid mode specification')

    def _index_closest(self, t, tol=None):
        d = np.abs(self - t)
        if tol is None:
            # If no tolerance is specified, use one clock tick of the
            # base_unit:
            tol = clock_tick

        # tolerance is converted into a time-array, so that it does the
        # right thing:
        ttol = TimeArray(tol, time_unit=self.time_unit)
        return np.where(d <= ttol)[0]

    def _index_before(self, t):
        # Use the standard Decorate-Sort-Undecorate (Schwartzian transform)
        # pattern to find the right index.
        cond = np.where(self <= t)[0]
        if len(cond) == 0:
            return cond
        idx_max = self[cond].argmax()
        return cond[idx_max]

    def _index_after(self, t):
        cond = np.where(t <= self)[0]
        if len(cond) == 0:
            return cond

        idx_min = self[cond].argmin()
        return cond[idx_min]

    def slice_during(self, e):
        """ Returns the slice that corresponds to Epoch e"""

        if not isinstance(e, Epochs):
            raise ValueError('e has to be of Epochs type')

        if e.data.ndim > 0:
            raise NotImplementedError('e has to be a scalar Epoch')

        if self.ndim != 1:
            e_s = 'slicing only implemented for 1-d TimeArrays'
            return NotImplementedError(e_s)

        # These two should be called with modes, such that they catch the right
        # slice
        start = self.index_at(e.start, mode='after')
        stop = self.index_at(e.stop, mode='before')

        # If *either* the start or stop index object comes back as the empty
        # array, then it means the condition is not satisfied, we return the
        # slice that does [:0],  i.e., always slices to nothing.
        if start.shape == (0,) or stop.shape == (0,):
            return slice(0)

        # Now,  we know the start/stop are not empty arrays, but they can be
        # either scalars or arrays.
        i_start = start if np.isscalar(start) else start.max()
        i_stop = stop if np.isscalar(stop) else stop.min()

        if e.start > self[i_start]:  # make sure self[i_start] is in epoch e
            i_start += 1
        if e.stop > self[i_stop]:  # make sure to include self[i_stop]
            i_stop += 1

        return slice(i_start, i_stop)

    def at(self, t, tol=None):
        """ Returns the values of the TimeArray object at time t"""
        return self[self.index_at(t, tol=tol)]

    def during(self, e):
        """ Returns the values of the TimeArray object during Epoch e"""

        if not isinstance(e, Epochs):
            raise ValueError('e has to be of Epochs type')

        if e.data.ndim > 0:
            ## TODO: Implement slicing with 1-d Epochs array,
            ## resulting in (ragged/jagged) 2-d TimeArray
            raise NotImplementedError('e has to be a scalar Epoch')

        return self[self.slice_during(e)]

##     def min(self,axis=None,out=None):
##         """Returns the minimal time"""
##         # this is a quick fix to return a time and will
##         # be obsolete once we use proper time dtypes
##         if axis is not None:
##             raise NotImplementedError, 'axis argument not implemented'
##         if out is not None:
##             raise NotImplementedError, 'out argument not implemented'
##         if self.ndim:
##             return self[self.argmin()]
##         else:
##             return self

    def max(self, axis=None, out=None):
        """Returns the maximal time"""
        # this is a quick fix to return a time and will
        # be obsolete once we use proper time dtypes
        if axis is not None:
            raise NotImplementedError('axis argument not implemented')
        if out is not None:
            raise NotImplementedError('out argument not implemented')
        if self.ndim:
            return self[self.argmax()]
        else:
            return self

    def convert_unit(self, time_unit):
        """Convert from one time unit to another in place"""

        self.time_unit = time_unit
        self._conversion_factor = time_unit_conversion[time_unit]

    def __div__(self, d):
        """Division by another time object eliminates units """
        if isinstance(d, TimeInterface):
            return np.divide(np.array(self), np.array(d).astype(float))
        else:
            return np.divide(self, d)

    __truediv__ = __div__ # called by python3

# Globally define a single tick of the base unit:
clock_tick = TimeArray(1, time_unit=base_unit)


class UniformTime(np.ndarray, TimeInterface):
    """ A representation of time sampled uniformly
    """

    def __new__(cls, data=None, length=None, duration=None, sampling_rate=None,
                sampling_interval=None, t0=0, time_unit=None):
        """

        Parameters
        ----------
        length : int
            The number of items in the time-array

        duration : float,
            the duration to be represented (given in the time-unit) of the
            array. If this item is an TimeArray, the units of the UniformTime
            array resulting will 'inherit' the units of the
            duration. Otherwise, the unit of the UniformTime will be set by
            that kwarg

        sampling_rate : float
            The sampling rate (in Hz)

        sampling_interval : float
            The inverse of the sampling_interval

        t0 : float, int or singleton `TimeArray`
            The value of the first time-point in the array (unless given as a
            `TimeArray`, should be in the time-unit)

        time_unit : str, optional
            The time unit to be used in the representation of time

        """

        # Sanity checks. There are different valid combinations of inputs
        tspec = tuple(x is not None for x in
                      [sampling_interval, sampling_rate, length, duration])

        # Used in converting tspecs to human readable form
        tspec_arg_names = ['sampling_interval',
                           'sampling_rate',
                           'length',
                           'duration']

        # The valid configurations
        valid_tspecs = [
            # interval, length:
            (True, False, True, False),
            # interval, duration:
            (True, False, False, True),
            # rate, length:
            (False, True, True, False),
            # rate, duration:
            (False, True, False, True),
            # length, duration:
            (False, False, True, True)
            ]

        if isinstance(data, UniformTime):
            # Assuming data was given, some other tspecs become valid:
            tspecs_w_data = dict(
                    nothing=(False, False, False, False),
                    sampling_interval=(True, False, False, False),
                    sampling_rate=(False, True, False, False),
                    length=(False, False, True, False),
                    duration=(False, False, False, True))
            # preserve the order of the keys
            valid_tspecs.append(tspecs_w_data['nothing'])
            for name in tspec_arg_names:
                valid_tspecs.append(tspecs_w_data[name])

        if (tspec not in valid_tspecs):
            # l = ['sampling_interval', 'sampling_rate', 'length', 'duration']
            # args = [arg for t,arg in zip(tspec,l) if t]
            raise ValueError("Invalid time specification.\n" +
                "You provided: %s \n"
                "%s \nsee docstring for more info."
                % (str_tspec(tspec, tspec_arg_names),
                  str_valid_tspecs(valid_tspecs,
                                   tspec_arg_names)))

        if isinstance(data, UniformTime):
            # Get attributes from the UniformTime object and transfer those
            # over:
            if tspec == tspecs_w_data['nothing']:
                sampling_rate = data.sampling_rate
                duration = data.duration
            elif tspec == tspecs_w_data['sampling_interval']:
                duration == data.duration
            elif tspec == tspecs_w_data['sampling_rate']:
                if isinstance(sampling_rate, Frequency):
                    sampling_interval = sampling_rate.to_period()
                else:
                    sampling_interval = 1.0 / sampling_rate
                duration = data.duration
            elif tspec == tspecs_w_data['length']:
                duration = length * data.sampling_interval
                sampling_rate = data.sampling_rate
            elif tspec == tspecs_w_data['duration']:
                sampling_rate = data.sampling_rate
            if time_unit is None:
                # If the user didn't ask to change the time-unit, use the
                # time-unit from the object you got:
                time_unit = data.time_unit

        # Check that the time units provided are sensible:
        if time_unit not in time_unit_conversion:
            raise ValueError('Invalid time unit %s, must be one of %s' %
                         (time_unit, time_unit_conversion.keys()))

        # Make sure you have a time unit:
        if time_unit is None:
            #If you gave us a duration with time_unit attached
            if isinstance(duration, TimeInterface):
                time_unit = duration.time_unit
            #Otherwise, you might have given us a sampling_interval with a
            #time_unit attached:
            elif isinstance(sampling_interval, TimeInterface):
                time_unit = sampling_interval.time_unit
            else:
                time_unit = 's'

        # Calculate the sampling_interval or sampling_rate:
        if sampling_interval is None:
            if isinstance(sampling_rate, Frequency):
                c_f = time_unit_conversion[time_unit]
                sampling_interval = sampling_rate.to_period() / float(c_f)
            elif sampling_rate is None:
                sampling_interval = float(duration) / length
                sampling_rate = Frequency(1.0 / sampling_interval,
                                          time_unit=time_unit)
            else:
                c_f = time_unit_conversion[time_unit]
                sampling_rate = Frequency(sampling_rate, time_unit='s')
                sampling_interval = sampling_rate.to_period() / float(c_f)
        else:
            if isinstance(sampling_interval, TimeInterface):
                c_f = time_unit_conversion[sampling_interval.time_unit]
                sampling_rate = Frequency(1.0 / (float(sampling_interval) /
                                                                       c_f),
                                     time_unit=sampling_interval.time_unit)
            else:
                sampling_rate = Frequency(1.0 / sampling_interval,
                                          time_unit=time_unit)

        # Calculate the duration, if that is not defined:
        if duration is None:
            duration = length * sampling_interval

        # 'cast' the time inputs as TimeArray
        duration = TimeArray(duration, time_unit=time_unit)
        #XXX If data is given - the t0 should be taken from there:
        t0 = TimeArray(t0, time_unit=time_unit)
        sampling_interval = TimeArray(sampling_interval, time_unit=time_unit)

        # in order for time[-1]-time[0]==duration to be true (which it should)
        # add the sampling_interval to the stop value:
        # time = np.arange(np.int64(t0),
        #                  np.int64(t0+duration+sampling_interval),
        #                  np.int64(sampling_interval),dtype=np.int64)

        # But it's unclear whether that's really the behavior we want?
        time = np.arange(np.int64(t0), np.int64(t0 + duration),
                         np.int64(sampling_interval), dtype=np.int64)

        time = np.asarray(time).view(cls)
        time.time_unit = time_unit
        time._conversion_factor = time_unit_conversion[time_unit]
        time.duration = duration
        time.sampling_rate = Frequency(sampling_rate)
        time.sampling_interval = sampling_interval
        time.t0 = t0

        return time

    def __array_wrap__(self, out_arr, context=None, return_scalar=False):
        # When doing comparisons between UniformTime, make sure that you return
        # a boolean array, not a time array:
        if out_arr.dtype == bool:
            ret = np.asarray(out_arr)
            if return_scalar:
                ret = ret[()]
            return ret
        else:
            return np.ndarray.__array_wrap__(self, out_arr, context, return_scalar)

    def __array_finalize__(self, obj):
        """XXX """
        # Make sure that the UniformTime has the time units set (and not equal
        # to None):
        if not hasattr(self, 'time_unit') or self.time_unit is None:
            if hasattr(obj, 'time_unit'):  # looks like view cast
                self.time_unit = obj.time_unit
            else:
                self.time_unit = 's'

        # Make sure that the conversion factor is set properly:
        if not hasattr(self, '_conversion_factor'):
            if hasattr(obj, '_conversion_factor'):
                self._conversion_factor = obj._conversion_factor
            else:
                self._conversion_factor = time_unit_conversion[self.time_unit]

        # Make sure that t0 attribute is set properly:
        for attr in ['t0', 'sampling_rate', 'sampling_interval', 'duration']:
            if not hasattr(self, attr) and hasattr(obj, attr):
                setattr(self, attr, getattr(obj, attr))

    def __repr__(self):
        """Pass it through the conversion factor"""

        #If the input is a single int/float (with no shape) return a 'scalar'
        #time-point:
        if self.shape == ():
            return "%r %s" % (int(self) / float(self._conversion_factor),
                            self.time_unit)

        #Otherwise, return the UniformTime representation:
        else:
            return np.ndarray.__repr__(self / float(self._conversion_factor)
             )[:-1] + ", time_unit='%s')" % self.time_unit

    def __getitem__(self, key):
        # return scalar TimeArray in case key is integer
        if isinstance(key, (int, np.int64, np.int32)):
            return self[[key]].reshape(()).view(TimeArray)
        elif isinstance(key, float) or isinstance(key, TimeInterface):
            return self.at(key)
        elif isinstance(key, Epochs):
            return self.during(key)
        else:
            return np.ndarray.__getitem__(self, key)

    def __setitem__(self, key, val):
        raise ValueError("""Setting of individual indices would break uniformity:
            You can either use += on the full array, OR
            create a new TimeArray from this UniformTime""")

    def _convert_and_check_uniformity(self, val):
        # look at the units - convert the values to what they need to be (in
        # the base_unit) and then delegate to the ndarray.__iadd__
        if not hasattr(val, '_conversion_factor'):
            val = np.asarray(val)
            if getattr(val, 'dtype', None) == np.int32:
                # we'll overflow if val's dtype is np.int32
                val = np.array(val, dtype=np.int64)
            val *= self._conversion_factor
        if hasattr(val, 'ndim') and val.ndim == 1:
            # we have to check that adding this will preserve uniformity
            dv = np.diff(val)
            uniformity_breaks, = np.where(dv!=dv[0])
            if len(uniformity_breaks) != 0:
                raise ValueError(
                    """All elements in the operand array must have a constant
                    interval between them in order to preserve uniformity.
                    Uniformity is broken at these indices: %s
                    """ %str(uniformity_breaks))
            self.sampling_interval += dv[0]
            self.sampling_rate = Frequency(1.0 / (float(self.sampling_interval) /
                                        time_unit_conversion[self.time_unit]),
                                        time_unit=self.time_unit)
        return val

    def __iadd__(self, val):
        val = self._convert_and_check_uniformity(val)
        return np.ndarray.__iadd__(self, val)

    def __isub__(self, val):
        val = self._convert_and_check_uniformity(val)
        return np.ndarray.__isub__(self, val)

    def __imul__(self, val):
        np.ndarray.__imul__(self, val)
        self.sampling_interval *= val
        self.sampling_rate = Frequency(self.sampling_rate / val)
        return self

    def __idiv__(self, val):
        np.ndarray.__idiv__(self, val)
        self.sampling_interval /= val
        self.sampling_rate = Frequency(self.sampling_rate * val)
        return self

    __itruediv__ =  __idiv__ # for py3k

    def index_at(self, t, boolean=False):
        """Find the index that corresponds to the time bin containing t

           Returns boolean mask if boolean=True and integer indices otherwise.
        """

        # cast t into time
        ta = TimeArray(t, time_unit=self.time_unit)

        # check that index is within range
        if ta.min() < self.t0 or ta.max() >= self.t0 + self.duration:
            raise ValueError('index out of range')
        idx = (ta - self.t0) // self.sampling_interval
        if boolean:
            bool_idx = np.zeros(len(self), dtype=bool)
            bool_idx[idx] = True
            return bool_idx
        elif ta.ndim == 0:
            return idx[()]
        else:
            return idx.view(np.ndarray)

    def slice_during(self, e):
        """ Returns the slice that corresponds to Epoch e"""

        if not isinstance(e, Epochs):
            raise ValueError('e has to be of Epochs type')

        if e.data.ndim > 0:
            raise NotImplementedError('e has to be a scalar Epoch')

        if self.ndim != 1:
            e_s = 'slicing only implemented for 1-d TimeArrays'
            return NotImplementedError(e_s)
        i_start = self.index_at(e.start)
        i_stop = self.index_at(e.stop)
        if e.start > self[i_start]:  # make sure self[i_start] is in epoch e
            i_start += 1
        if e.stop > self[i_stop]:  # make sure to include self[i_stop]
            i_stop += 1

        return slice(i_start, i_stop)

    def at(self, t):
        """ Returns the values of the UniformTime object at time t"""
        return TimeArray(self[self.index_at(t)], time_unit=self.time_unit)

    def during(self, e):
        """ Returns the values of the UniformTime object during Epoch e"""

        if not isinstance(e, Epochs):
            raise ValueError('e has to be of Epochs type')

        if e.data.ndim > 0:
            raise NotImplementedError('e has to be a scalar Epoch')

        return self[self.slice_during(e)]

    def min(self, axis=None, out=None):
        """Returns the minimal time"""
        # this is a quick fix to return a time and will
        # be obsolete once we use proper time dtypes
        if axis is not None:
            raise NotImplementedError('axis argument not implemented')
        if out is not None:
            raise NotImplementedError('out argument not implemented')
        if self.ndim:
            return self[self.argmin()]
        else:
            return self

    def max(self, axis=None, out=None):
        """Returns the maximal time"""
        # this is a quick fix to return a time and will
        # be obsolete once we use proper time dtypes
        if axis is not None:
            raise NotImplementedError('axis argument not implemented')
        if out is not None:
            raise NotImplementedError('out argument not implemented')
        if self.ndim:
            return self[self.argmax()]
        else:
            return self

    def __div__(self, d):
        """Division by another time object eliminates units """
        if isinstance(d, TimeInterface):
            return np.divide(np.array(self), np.array(d).astype(float))
        else:
            return np.divide(self, d)

    __truediv__ =  __div__ # for py3k

##Frequency:

class Frequency(float):
    """A class for representation of the frequency (in Hz) """

    def __new__(cls, f, time_unit='s'):
        """Initialize a frequency object """

        tuc = time_unit_conversion
        scale_factor = (float(tuc['s']) / tuc[time_unit])
        #If the input is a Frequency object, it is already in Hz:
        if isinstance(f, Frequency) == False:
            #But otherwise convert to Hz:
            f = f * scale_factor

        freq = super(Frequency, cls).__new__(cls, f)
        freq._time_unit = time_unit

        return freq

    def __repr__(self):

        return str(float(self)) + ' Hz'


    def to_period(self, time_unit=base_unit):
        """Convert the value of a frequency to the corresponding period
        (defaulting to a representation in the base_unit)

        """
        tuc = time_unit_conversion
        scale_factor = (float(tuc['s']) / tuc[time_unit])

        return np.int64((1 / self) * scale_factor)


##Time-series:
class TimeSeriesInterface(TimeInterface):
    """The minimally agreed upon interface for all time series.

    This should be thought of as an abstract base class.
    """
    time = None
    data = None
    metadata = None


class TimeSeriesBase(object):
    """Base class for time series, implementing the TimeSeriesInterface."""

    def __init__(self, data, time_unit, metadata=None):
        """Common constructor shared by all TimeSeries classes."""
        # Check that sensible time units were given
        if time_unit not in time_unit_conversion:
            raise ValueError('Invalid time unit %s, must be one of %s' %
                             (time_unit, time_unit_conversion.keys()))

        #: the data is an arbitrary numpy array
        self.data = np.asanyarray(data)
        self.time_unit = time_unit

        # Every instance carries an empty metadata dict, which we promise never
        # to touch.  This reserves this name as a user area for extra
        # information without the danger of name clashes in the future.
        if metadata is None:
            self.metadata = {}
        else:
            self.metadata = metadata

    def __len__(self):
        """Return the length of the time series."""
        return self.data.shape[-1]

    def _validate_dimensionality(self):
        """Check that the data and time have the proper dimensions.
        """

        if self.time.ndim != 1:
            raise ValueError("time array must be one-dimensional")
        npoints = self.data.shape[-1]
        if npoints != len(self.time):
            raise ValueError("mismatch of time and data dimensions")

    def __getitem__(self, key):
        """use fancy time-indexing (at() method)."""
        if isinstance(key, TimeInterface):
            return self.at(key)
        elif isinstance(key, Epochs):
            return self.during(key)
        elif self.data.ndim == 1:
            return self.data[key]  # time is the last dimension
        else:
            return self.data[..., key]  # time is the last dimension

    def __repr__(self):
        rep = self.__class__.__name__ + ":"
        return rep + self.time.__repr__() + self.data.T.__repr__()

    # add some methods that implement arithmetic on the timeseries data
    def __add__(self, other):
        out = self.copy()
        out.data = out.data.__add__(np.asanyarray(other).T)
        return out

    def __sub__(self, other):
        out = self.copy()
        out.data = out.data.__sub__(np.asanyarray(other).T)
        return out

    def __mul__(self, other):
        out = self.copy()
        out.data = out.data.__mul__(np.asanyarray(other).T)
        return out

    def __div__(self, other):
        out = self.copy()
        out.data = out.data.__truediv__(np.asanyarray(other).T)
        return out

    __truediv__ = __div__  # for py3k

    def __iadd__(self, other):
        self.data.__iadd__(np.asanyarray(other).T)
        return self

    def __isub__(self, other):
        self.data.__isub__(np.asanyarray(other).T)
        return self

    def __imul__(self, other):
        self.data.__imul__(np.asanyarray(other).T)
        return self

    def __idiv__(self, other):
        self.data.__itruediv__(np.asanyarray(other).T)
        return self

    __itruediv__ = __idiv__  # for py3k

class TimeSeries(TimeSeriesBase):
    """Represent data collected at uniform intervals.
    """

    @desc.setattr_on_read
    def time(self):
        """Construct time array for the time-series object. This holds a
    UniformTime object, with properties derived from the TimeSeries
    object"""
        return UniformTime(length=self.__len__(), t0=self.t0,
                           sampling_interval=self.sampling_interval,
                           time_unit=self.time_unit)

    #XXX This should call the constructor in an appropriate way, when provided
    #with a UniformTime object and data, so that you don't need to deal with
    #the constructor itself:
    @staticmethod
    def from_time_and_data(time, data):
        return TimeSeries.__init__(data, time=time)

    def copy(self):
        return TimeSeries(data=self.data.copy(),
                          time=self.time.copy(),
                          time_unit=self.time_unit,
                          metadata=self.metadata.copy())

    def __init__(self, data, t0=None, sampling_interval=None,
                 sampling_rate=None, duration=None, time=None, time_unit='s',
                 metadata=None):
        """Create a new TimeSeries.

        This class assumes that data is uniformly sampled, but you can specify
        the sampling in one of three (mutually exclusive) ways:

        - sampling_interval [, t0]: data sampled starting at t0, equal
          intervals of sampling_interval.

        - sampling_rate [, t0]: data sampled starting at t0, equal intervals of
          width 1/sampling_rate.

        - time: a UniformTime object, in which case the TimeSeries can
          'inherit' the properties of this object.

        Parameters
        ----------
        data : array_like
          Data array, interpreted as having its last dimension being time.
        sampling_interval : float
          Interval between successive time points.
        sampling_rate : float
          Inverse of the interval between successive time points.
        t0 : float
          If you provide a sampling rate, you can optionally also provide a
          starting time.
        time
          Instead of sampling rate, you can explicitly provide an object of
          class UniformTime. Note that you can still also provide a different
          sampling_rate/sampling_interval/duration to take the place of the
          one in this object, but only as long as the changes are consistent
          with the length of the data.

        time_unit :  string
          The unit of time.

        Examples
        --------

        The minimal specification of data and sampling interval:

        >>> ts = TimeSeries([1,2,3],sampling_interval=0.25)
        >>> ts.time
        UniformTime([0.  , 0.25, 0.5 ], time_unit='s')
        >>> ts.t0
        0.0 s
        >>> ts.sampling_rate
        4.0 Hz

        Or data and sampling rate:

        >>> ts = TimeSeries([1,2,3],sampling_rate=2)
        >>> ts.time
        UniformTime([0. , 0.5, 1. ], time_unit='s')
        >>> ts.t0
        0.0 s
        >>> ts.sampling_interval
        0.5 s

        A time series where we specify the start time and sampling interval:

        >>> ts = TimeSeries([1,2,3],t0=4.25,sampling_interval=0.5)
        >>> ts.data
        array([1, 2, 3])
        >>> ts.time
        UniformTime([4.25, 4.75, 5.25], time_unit='s')
        >>> ts.t0
        4.25 s
        >>> ts.sampling_interval
        0.5 s
        >>> ts.sampling_rate
        2.0 Hz

        >>> ts = TimeSeries([1,2,3],t0=4.25,sampling_rate=2.0)
        >>> ts.data
        array([1, 2, 3])
        >>> ts.time
        UniformTime([4.25, 4.75, 5.25], time_unit='s')
        >>> ts.t0
        4.25 s
        >>> ts.sampling_interval
        0.5 s
        >>> ts.sampling_rate
        2.0 Hz

        """

        #If a UniformTime object was provided as input:
        if isinstance(time, UniformTime):
            c_fac = time._conversion_factor
            #If the user did not provide an alternative t0, get that from the
            #input:
            if t0 is None:
                t0 = time.t0
            #If the user did not provide an alternative sampling interval/rate:
            if sampling_interval is None and sampling_rate is None:
                sampling_interval = time.sampling_interval
                sampling_rate = time.sampling_rate
            #The duration can be read either from the length of the data, or
            #from the duration specified by the time-series:
            if duration is None:
                duration = time.duration
                length = time.shape[-1]
                #If changing the duration requires a change to the
                #sampling_rate, make sure that this was explicitly required by
                #the user - if the user did not explicitly set the
                #sampling_rate, or it is inconsistent, throw an error:
                data_len = np.array(data).shape[-1]

                if (length != data_len and
                    sampling_rate != float(data_len * c_fac) / time.duration):
                    e_s = "Length of the data (%s) " % str(len(data))
                    e_s += "specified sampling_rate (%s) " % str(sampling_rate)
                    e_s += "do not match."
                    raise ValueError(e_s)
            #If user does not provide a
            if time_unit is None:
                time_unit = time.time_unit

        else:
            ##If the input was not a UniformTime, we need to check that there
            ##is enough information in the input to generate the UniformTime
            ##array.

            #There are different valid combinations of inputs
            tspec = tuple(x is not None for x in
                      [sampling_interval, sampling_rate, duration])

            tspec_arg_names = ["sampling_interval",
                               "sampling_rate",
                               "duration"]

            #The valid configurations
            valid_tspecs = [
                      #interval, length:
                      (True, False, False),
                      #interval, duration:
                      (True, False, True),
                      #rate, length:
                      (False, True, False),
                      #rate, duration:
                      (False, True, True),
                      #length, duration:
                      (False, False, True)
                      ]

            if tspec not in valid_tspecs:
                raise ValueError("Invalid time specification. \n"
                      "You provided: %s\n %s see docstring for more info." % (
                            str_tspec(tspec, tspec_arg_names),
                            str_valid_tspecs(valid_tspecs, tspec_arg_names)))

        # Make sure to grab the time unit from the inputs, if it is provided:
        if time_unit is None:
            # If you gave us a duration with time_unit attached
            if isinstance(duration, TimeInterface):
                time_unit = duration.time_unit
            # Otherwise, you might have given us a sampling_interval with a
            # time_unit attached:
            elif isinstance(sampling_interval, TimeInterface):
                time_unit = sampling_interval.time_unit

        # Calculate the sampling_interval or sampling_rate from each other and
        # assign t0, if it is not already assigned:
        if sampling_interval is None:
            if isinstance(sampling_rate, Frequency):
                c_f = time_unit_conversion[time_unit]
                sampling_interval = sampling_rate.to_period() / float(c_f)
            elif sampling_rate is None:
                data_len = np.asarray(data).shape[-1]
                sampling_interval = float(duration) / data_len
                sampling_rate = Frequency(1.0 / sampling_interval,
                                             time_unit=time_unit)
            else:
                c_f = time_unit_conversion[time_unit]
                sampling_rate = Frequency(sampling_rate, time_unit='s')
                sampling_interval = sampling_rate.to_period() / float(c_f)
        else:
            if sampling_rate is None:  # Only if you didn't already 'inherit'
                                       # this property from another time object
                                       # above:
                if isinstance(sampling_interval, TimeInterface):
                    c_f = time_unit_conversion[sampling_interval.time_unit]
                    sampling_rate = Frequency(1.0 / (float(sampling_interval) /
                                                                         c_f),
                                       time_unit=sampling_interval.time_unit)
                else:
                    sampling_rate = Frequency(1.0 / sampling_interval,
                                              time_unit=time_unit)

        #Calculate the duration, if that is not defined:
        if duration is None:
            duration = np.asarray(data).shape[-1] * sampling_interval

        if t0 is None:
            t0 = 0

        # Make sure to grab the time unit from the inputs, if it is provided:
        if time_unit is None:
            #If you gave us a duration with time_unit attached
            if isinstance(duration, TimeInterface):
                time_unit = duration.time_unit
            #Otherwise, you might have given us a sampling_interval with a
            #time_unit attached:
            elif isinstance(sampling_interval, TimeInterface):
                time_unit = sampling_interval.time_unit

        #Otherwise, you can still call the common constructor to get the real
        #object initialized, with time_unit set to None and that will generate
        #the object with time_unit set to 's':
        TimeSeriesBase.__init__(self, data, time_unit, metadata=metadata)

        self.time_unit = time_unit
        self.sampling_interval = TimeArray(sampling_interval,
                                           time_unit=self.time_unit)
        self.t0 = TimeArray(t0, time_unit=self.time_unit)
        self.sampling_rate = sampling_rate
        self.duration = TimeArray(duration, time_unit=self.time_unit)

    def at(self, t, tol=None):
        """ Returns the values of the TimeArray object at time t"""
        return self.data[..., self.time.index_at(t)]

    def during(self, e):
        """ Returns the TimeSeries slice corresponding to epoch e """

        if not isinstance(e, Epochs):
            raise ValueError('e has to be of Epochs type')

        if e.data.ndim == 0:
            return TimeSeries(data=self.data[..., self.time.slice_during(e)],
                              time_unit=self.time_unit, t0=e.offset,
                              sampling_rate=self.sampling_rate)
        else:
            # TODO: make this a more efficient implementation, naive first pass
            if (e.duration != e.duration[0]).any():
                raise ValueError("All epochs must have the same duration")

            data = np.array([self.data[..., self.time.slice_during(ep)]
                             for ep in e])

            return TimeSeries(data=data,
                              time_unit=self.time_unit, t0=e.offset,
                              sampling_rate=self.sampling_rate)

    @property
    def shape(self):
        return self.data.shape


_epochtype = np.dtype({'names': ['start', 'stop'], 'formats': [np.int64] * 2})


class Epochs(desc.ResetMixin):
    """Represents a time interval"""
    def __init__(self, t0=None, stop=None, offset=None, start=None,
                 duration=None, time_unit=None, static=None, **kwargs):
        """
        Parameters
        ----------
        t0 : 1-d array or `TimeArray`
           A time relative to which the epochs started. Per default `t0` and
          `start` are the same, but setting the `offset` parameter can adjust
           that, so that the start-times are at a fixed time, relative to t0.

        stop : 1-d array or `TimeArray`
              The times of ends of epochs

        offset : float, int or singleton `TimeArray`
            A constant offset applied to t0 to set the starts of Epochs

        start : 1-d array or `TimeArray`
              The times of beginnings of epochs

        duration : 1-d array or `TimeArray`
           The durations of intervals.

        time_unit : str, optional
              The time unit of the object and all time-related things in it.
              Default: 's'

        static : dict, optional
            For fast initialization of an `Epochs` object from another `Epochs`
            object, this dict should contain all necessary items to have an
            `Epoch` defined.

        """
        # Short-circuit path for a fast initialization. This relies on `static`
        # to be a dict that contains everything that defines an Epochs class
        # XXX: add this sort of fast __init__ to all other classes
        if static is not None:
            self.__dict__.update(static)
            # we have to reset the duration OneTimeProperty, since it refers
            # to computations performed on the former object
            self.reset()
            return

        if t0 is None and start is None:
            raise ValueError('Either start or t0 need to be specified')
        # Normal, error checking and type converting initialization logic

        if stop is None and duration is None:
            raise ValueError('Either stop or duration have to be specified')

        if stop is not None and duration is not None:
            ### TODO: check if stop and duration are consistent
            e_s = 'Only either stop or duration have to be specified'
            raise ValueError(e_s)

        if offset is None:
            offset = 0

        t_offset = TimeArray(offset, time_unit=time_unit)

        if t_offset.ndim > 0:
            raise ValueError('Only scalar offset allowed')

        if t0 is None:
            t_0 = 0
        else:
            t_0 = TimeArray(t0, time_unit=time_unit)

        if start is None:
            t_start = t_0 - t_offset
        else:
            t_start = TimeArray(start, time_unit=time_unit)

        # inherit time_unit of t_start
        self.time_unit = t_start.time_unit

        if stop is None:
            t_duration = TimeArray(duration, time_unit=time_unit)
            t_stop = t_start + t_duration
        else:
            t_stop = TimeArray(stop, time_unit=time_unit)

        if t_start.shape != t_stop.shape:
            raise ValueError('start and stop have to have same shape')

        if t_start.ndim == 0:
            # return a 'scalar' epoch
            self.data = np.empty(1, dtype=_epochtype).reshape(())
        elif t_start.ndim == 1:
            # return a 1-d epoch array
            self.data = np.empty(t_start.shape[0], dtype=_epochtype)
        else:
            e_s = 'Only 0-dim and 1-dim start and stop times allowed'
            raise ValueError(e_s)

        self.data['start'] = t_start
        self.data['stop'] = t_stop

        self.offset = t_offset

    # TODO: define setters for start, stop, offset attributes
    @property
    def start(self):
        return TimeArray(self.data['start'],
                         time_unit=self.time_unit,
                         copy=False)

    @property
    def stop(self):
        return TimeArray(self.data['stop'],
                         time_unit=self.time_unit,
                         copy=False)

    @desc.setattr_on_read
    def duration(self):
        """Duration array for the epoch"""
        return self.stop - self.start

    def __getitem__(self, key):
        # create the static dict needed for fast version of __init__
        static = self.__dict__.copy()
        static['data'] = self.data[key]
        # self.__class__ here is Epochs or a subclass of Epochs
        # and `start` is a required argument
        return self.__class__(start=None, static=static)

    def __repr__(self):
        if self.data.ndim == 0:
            z = (self.start, self.stop)
        else:
            z = list(zip(self.start, self.stop))
        rep = self.__class__.__name__ + "(" + z.__repr__()
        return rep + ", as (start,stop) tuples)"

    def __len__(self):
        return len(self.data)


def str_tspec(tspec, arg_names):
    """ Turn a single tspec into human readable form"""
    # an all "False" will convert to an empty string unless we do the following
    # where we create an all False tuple of the appropriate length
    if tspec == tuple([False] * len(arg_names)):
        return "(nothing)"
    return ", ".join([arg for t, arg in zip(tspec, arg_names) if t])


def str_valid_tspecs(valid_tspecs, arg_names):
    """Given a set of valid_tspecs, return a string that turns them into
    human-readable form"""
    vargs = []
    for tsp in valid_tspecs:
        vargs.append(str_tspec(tsp, arg_names))
    return "\n Valid time specifications are:\n\t%s" % ("\n\t".join(vargs))


def concatenate_time_series(time_series_seq):
    """Concatenates a sequence of time-series objects in time.

    The input can be any iterable of time-series objects; metadata, sampling
    rates and other attributes are kept from the last one in the sequence.

    This one requires that all the time-series in the list have the same
    sampling rate and that all the data have the same number of items in all
    dimensions, except the time dimension"""

    # Extract the data pointer for each and build a common data block
    data = []
    metadata = {}
    for ts in time_series_seq:
        data.append(ts.data)
        metadata.update(ts.metadata)

    # Sampling interval is read from the last one
    tseries = TimeSeries(np.concatenate(data,-1),
                                sampling_interval=ts.sampling_interval,
                                metadata=metadata)
    return tseries


class Events(TimeInterface):
    """Represents timestamps and associated data """

    def __init__(self, time, labels=None, indices=None,
                 time_unit=None, **data):
        """
        Parameters
        ----------
        time : array or TimeArray
            The times at which events occurred

        labels : array, optional

        indices : int array, optional


        Notes
        -----


        """
        # The time data must be at least a 1-d array, NOT a time scalar
        if not np.iterable(time):
            time = [time]

        # First initialize the TimeArray from the time-stamps
        self.time = TimeArray(time, time_unit=time_unit)
        self.time_unit = self.time.time_unit

        # Make sure time is one-dimensional
        if self.time.ndim != 1:
            e_s = 'The TimeArray provided can only be one-dimensional'
            raise ValueError(e_s)
        # Ensure that the dict of data values has a known, uniform structure:
        # all values must be arrays, with at least one dimension.
        new_data = {}
        for k, v in data.items():
            if np.iterable(v):
                v = np.asanyarray(v)
            else:
                # For scalars, we do NOT want to create 0-d arrays, which are
                # rather tricky to work with.  So if the input value is not an
                # iterable object, we turn it into a one-element 1-d array.
                v = np.array([v])
            new_data[k] = v

        # Make sure all data has same length
        ntimepts = len(self.time)
        for check_v in new_data.values():
            if len(check_v) != ntimepts:
                e_s = 'All data in the Events must be of the same'
                e_s += 'length as the associated time'
                raise ValueError(e_s)

        # Make sure indices have same length and are integers
        if labels is not None:
            if len(labels) != len(indices):
                e_s = 'Labels and indices must have the same length'
                raise ValueError(e_s)
            dt = [(l, np.int64) for l in labels]
        else:
            dt = np.int64
            dt = [('i%d' % i, np.int64)
                  for i in range(len(indices or ()))] or np.int64

        self.index = np.array(list(zip(*(indices or ()))),
                                       dtype=dt).view(np.recarray)

        #Should data be a recarray?
##         dt = [(st,np.array(data[st]).dtype) for st in data] or None
##         self.data = np.array(zip(*data.values()),
##         dtype=dt).view(np.recarray)

        #Or a dict?
        self.data = new_data

    def __repr__(self):
        rep = self.__class__.__name__ + ":\n\t"
        rep += repr(self.time) + "\n\t"
        rep += repr(self.data)
        return rep

    def __getitem__(self, key):
        # return scalar TimeArray in case key is integer
        newdata = dict()
        newtime = self.time[key].reshape(-1)
        sl = key
        if isinstance(key, float):
            sl = self.time.index_at(key)
        elif isinstance(key, Epochs):
            sl = self.time.slice_during(key)
        for k, v in self.data.items():
            newdata[k] = v[sl]

        # XXX: I don't really understand how labels and index are supposed to
        # be used, so I'm not implementing them when slicing events - pi
        # 2010-12-04

        # self.__class__ here is Events or a subclass of Events
        return self.__class__(newtime, **newdata)

    def __len__(self):
        return len(self.time)



import sys
import numpy as np
import numpy.testing as npt

import pytest

import nitime.lazyimports as l

# The next test requires nitime.lazyimports.disable_lazy_imports to have been
# set to false, otherwise the lazy import machinery is disabled and all imports
# happen at l.LazyImport calls which become equivalent to regular import
# statements
@pytest.mark.skipif(l.disable_lazy_imports, reason="Lazy imports disabled")
def test_lazy():
    mlab = l.LazyImport('matplotlib.mlab')
    # repr for mlab should be <module 'matplotlib.mlab' will be lazily loaded>
    assert 'lazily loaded' in repr(mlab)
    # accessing mlab's attribute will cause an import of mlab
    npt.assert_(np.all(mlab.detrend_mean(np.array([1, 2, 3]))  ==
                       np.array([-1., 0., 1.])))
    # now mlab should be of class LoadedLazyImport an repr(mlab) should be
    # <module 'matplotlib.mlab' from # '.../matplotlib/mlab.pyc>
    assert 'lazily loaded' not in repr(mlab)

# A known limitation of our lazy loading implementation is that, when it it is
# enabled, reloading the module raises an ImportError, and it also does not
# actually perform a reload, as demonstrated by this test.
@pytest.mark.skipif(l.disable_lazy_imports, reason="Lazy imports disabled")
def test_lazy_noreload():
    "Reloading of lazy modules causes ImportError"
    mod = l.LazyImport('sys')
    # accessing module dictionary will trigger an import
    len(mod.__dict__)
    # do not use named tuple feature for Python 2.6 compatibility
    major, minor = sys.version_info[:2]
    if major == 2:
        with pytest.raises(ImportError) as e_info:
            reload(mod)
    elif major == 3:
        import importlib
        with pytest.raises(ImportError) as e_info:
            importlib.reload(mod)



import numpy as np


def boxcar_filter(time_series, lb=0, ub=0.5, n_iterations=2):
    """
    Filters data into a frequency range.

    For each of the two bounds, a low-passed version is created by convolving
    with a box-car and then the low-passed version for the upper bound is added
    to the low-passed version for the lower bound subtracted from the signal,
    resulting in a band-passed version

    Parameters
    ----------

    time_series: float array
       the signal
    ub : float, optional
      The cut-off frequency for the low-pass filtering as a proportion of the
      sampling rate. Default to 0.5 (Nyquist)
    lb : float, optional
      The cut-off frequency for the high-pass filtering as a proportion of the
      sampling rate. Default to 0
    n_iterations: int, optional
      how many rounds of smoothing to do. Default to 2.

    Returns
    -------
    float array:
      The signal, filtered
    """

    n = time_series.shape[-1]

    len_boxcar_ub = np.ceil(1 / (2.0 * ub))
    boxcar_ub = np.empty(int(len_boxcar_ub))
    boxcar_ub.fill(1.0 / len_boxcar_ub)
    boxcar_ones_ub = np.ones_like(boxcar_ub)

    if lb == 0:
        lb = None
    else:
        len_boxcar_lb = np.ceil(1 / (2.0 * lb))
        boxcar_lb = np.empty(int(len_boxcar_lb))
        boxcar_lb.fill(1.0 / len_boxcar_lb)
        boxcar_ones_lb = np.ones_like(boxcar_lb)

    #If the time_series is a 1-d, we add a dimension, so that we can iterate
    #over 2-d inputs:
    if len(time_series.shape) == 1:
        time_series = np.array([time_series])
    for i in range(time_series.shape[0]):
        if ub:
            # Start by applying a low-pass to the signal.  Pad the signal on
            # each side with the initial and terminal signal value:
            pad_s = np.hstack((boxcar_ones_ub *
                               time_series[i, 0], time_series[i]))
            pad_s = np.hstack((pad_s, boxcar_ones_ub * time_series[i, -1]))

            # Filter operation is a convolution with the box-car(iterate,
            # n_iterations times over this operation):
            for iteration in range(n_iterations):
                conv_s = np.convolve(pad_s, boxcar_ub)

            # Extract the low pass signal by excising the central
            # len(time_series) points:
            time_series[i] = conv_s[conv_s.shape[-1] // 2 -
                                    int(np.floor(n / 2.)):
                                    conv_s.shape[-1] // 2 +
                                    int(np.ceil(n / 2.))]

        # Now, if there is a high-pass, do the same, but in the end subtract
        # out the low-passed signal:
        if lb:
            pad_s = np.hstack((boxcar_ones_lb *
                               time_series[i, 0], time_series[i]))
            pad_s = np.hstack((pad_s, boxcar_ones_lb * time_series[i, -1]))

            #Filter operation is a convolution with the box-car(iterate,
            #n_iterations times over this operation):
            for iteration in range(n_iterations):
                conv_s = np.convolve(pad_s, boxcar_lb)

            #Extract the low pass signal by excising the central
            #len(time_series) points:
            s_lp = conv_s[conv_s.shape[-1] // 2 - int(np.floor(n / 2.)):
                          conv_s.shape[-1] // 2 + int(np.ceil(n / 2.))]

            #Extract the high pass signal simply by subtracting the high pass
            #signal from the original signal:
            time_series[i] = time_series[i] - s_lp + np.mean(s_lp)  # add mean
            #to make sure that there are no negative values. This also seems to
            #make sure that the mean of the signal (in % signal change) is
            #close to 0

    return time_series.squeeze()



"""

Test the io submodule of the fMRI module of nitime

"""
import os

import numpy as np
import numpy.testing as npt
import pytest

import nitime
import nitime.timeseries as ts

#Skip the tests if you can't import nibabel:
try:
    import nitime.fmri.io as io
    no_nibabel = False
    no_nibabel_msg=''
except ImportError as e:
    no_nibabel = True
    no_nibabel_msg=e.args[0]

data_path = os.path.join(nitime.__path__[0],'data')

@pytest.mark.skipif(no_nibabel, reason=no_nibabel_msg)
def test_time_series_from_file():

    """Testing reading of data from nifti files, using nibabel"""

    TR = 1.35
    ts_ff = io.time_series_from_file

    #File names:
    fmri_file1 = os.path.join(data_path,'fmri1.nii.gz')
    fmri_file2 = os.path.join(data_path,'fmri2.nii.gz')

    #Spatial coordinates into the volumes:
    coords1 = np.array([[5,5,5,5],[5,5,5,5],[1,2,3,4]])
    coords2 = np.array([[6,6,6,6],[6,6,6,6],[3,4,5,6]])

    #No averaging, no normalization:
    t1 = ts_ff([fmri_file1,fmri_file2],[coords1,coords2],TR)

    npt.assert_equal(t1[0].shape,(4,80))  # 4 coordinates, 80 time-points

    t2 = ts_ff([fmri_file1,fmri_file2],[coords1,coords2],TR,average=True)

    npt.assert_equal(t2[0].shape,(80,))  # collapse coordinates,80 time-points

    t3 = ts_ff(fmri_file1,coords1,TR,normalize='zscore')

    #The mean of each channel should be almost equal to 0:
    npt.assert_almost_equal(t3.data[0].mean(),0)
    #And the standard deviation should be almost equal to 1:
    npt.assert_almost_equal(t3.data[0].std(),1)

    t4 = ts_ff(fmri_file1,coords1,TR,normalize='percent')

    #In this case, the average is almost equal to 0, but no constraint on the
    #std:
    npt.assert_almost_equal(t4.data[0].mean(),0)

    #Make sure that we didn't mess up the sampling interval:
    npt.assert_equal(t4.sampling_interval,nitime.TimeArray(1.35))

    # Test the default behavior:
    data = io.load(fmri_file1).get_fdata()
    t5 = ts_ff(fmri_file1)
    npt.assert_equal(t5.shape, data.shape)
    npt.assert_equal(t5.sampling_interval, ts.TimeArray(1, time_unit='s'))

    # Test initializing TR with a TimeArray:
    t6= ts_ff(fmri_file1, TR=ts.TimeArray(1350, time_unit='ms'))
    npt.assert_equal(t4.sampling_interval, t6.sampling_interval)

    # Check the concatenation dimensions:
    t7 = ts_ff([fmri_file1, fmri_file2])
    npt.assert_equal([t7.shape[:3], t7.shape[-1]], [data.shape[:3], data.shape[-1]*2])

    t8 = ts_ff([fmri_file1, fmri_file2], average=True)
    npt.assert_equal(t8.shape[0], data.shape[-1]*2)

    t9 = ts_ff([fmri_file1, fmri_file2], average=True, normalize='zscore')
    npt.assert_almost_equal(t9.data.mean(), 0)



import numpy as np
import numpy.testing as npt

import nitime.algorithms as tsa

np.random.seed(1945)

def test_entropy():
    x = np.random.randint(0, 2, size=1000)
    e1 = tsa.entropy(x)
    npt.assert_almost_equal(e1, 1, decimal=2)
    # The joint entropy of the variable with itself is the same:
    e2 = tsa.entropy(x, x)
    npt.assert_almost_equal(e1, e2)
    y = np.random.randint(0, 2, size=1000)
    # Joint entropy with another random variable is 2:
    e3 = tsa.entropy(x, y)
    npt.assert_almost_equal(e3, 2, decimal=2)


def test_conditional_entropy():
    x = np.random.randint(0, 2, size=1000)
    y = np.random.randint(0, 2, size=1000)
    e1 = tsa.conditional_entropy(x, x)
    npt.assert_almost_equal(e1, 0)
    e2 = tsa.conditional_entropy(x, y)
    npt.assert_almost_equal(e2, 1, decimal=2)


def test_mutual_information():
    x = np.random.randint(0, 2, size=1000)
    y = np.random.randint(0, 2, size=1000)
    e1 = tsa.mutual_information(x, x)
    npt.assert_almost_equal(e1, 1, decimal=2)
    e2 = tsa.mutual_information(x, y)
    npt.assert_almost_equal(e2, 0, decimal=2)


def test_entropy_cc():
    x = np.random.randint(0, 2, size=1000)
    e1 = tsa.entropy_cc(x, x)
    npt.assert_almost_equal(e1, 1, decimal=2)


def test_transfer_entropy():
    x = np.random.randint(0, 4, size=1000)
    y = np.roll(x, -1)
    e1 = tsa.transfer_entropy(x, y, lag=1)
    npt.assert_almost_equal(e1, 2, decimal=1)



#!/usr/bin/env python
"""Run the py->rst conversion and run all examples.

This also creates the index.rst file appropriately, makes figures, etc.
"""
# Library imports

# Stdlib imports
import os
import sys

from glob import glob

# Third-party imports

# We must configure the mpl backend before making any further mpl imports
import matplotlib
import matplotlib.pyplot as plt
from matplotlib._pylab_helpers import Gcf

# Local tools
from toollib import *

# Set MPL backend:
matplotlib.use('Agg')

# Globals:

examples_header = """

.. _examples:

========
Examples
========

.. include:: note_about_examples.txt

.. toctree::
   :maxdepth: 2


"""
# Function definitions::

# These global variables let show() be called by the scripts in the usual
# manner, but when generating examples, we override it to write the figures to
# files with a known name (derived from the script name) plus a counter
figure_basename = None

# We must change the show command to save instead
def show():
    """
    This over-rides matplotlibs `show` function to save instead of rendering to
    the screen.
    """
    allfm = Gcf.get_all_fig_managers()
    for fcount, fm in enumerate(allfm):
        fm.canvas.figure.savefig('%s_%02i.png' %
                                 (figure_basename, fcount+1))

_mpl_show = plt.show
plt.show = show

# Main script::

# Work in examples directory
cd('examples')
if not os.getcwd().endswith('doc/examples'):
    raise OSError('This must be run from doc/examples directory')

# Run the conversion from .py to rst file
sh('../../tools/ex2rst --project Nitime --outdir . .')

# Make the index.rst file
index = open('index.rst', 'w')
index.write(examples_header)
for name in [os.path.splitext(f)[0] for f in glob('*.rst')]:
    #  Don't add the index in there to avoid sphinx errors and don't add the
    #  note_about examples again (because it was added at the top):
    if name not in(['index', 'note_about_examples']):
        index.write('   %s\n' % name)
index.close()
# Execute each python script in the directory.
if '--no-exec' in sys.argv:
    pass
else:
    if not os.path.isdir('fig'):
        os.mkdir('fig')

    for script in glob('*.py'):
        print(script)
        figure_basename = pjoin('fig', os.path.splitext(script)[0])

        with open(script) as f:
            exec(f.read())
        plt.close('all')
